import json
import logging
from datetime import datetime

from requests.exceptions import HTTPError

from irida_sistr_results.SistrResultsException import SistrResultsException
from irida_sistr_results.sistr_info import SampleSistrInfo

LOGLEVEL_TRACE = 5

logger = logging.getLogger("irida-api")


class IridaAPI(object):
    """A class for dealing with higher-level API functionality of the IRIDA REST API."""

    def __init__(self, irida_connector):
        self.irida_connector = irida_connector

    def _get_rel_from_links(self, rel, links):
        href = None

        for l in links:
            if (l['rel'] == rel):
                href = l['href']

        if (href is None):
            raise Exception("Could not get rel=" + rel + " from links")
        else:
            return href

    def _has_rel_in_links(self, rel, links):
        for l in links:
            if (l['rel'] == rel):
                return True
        return False

    def _log_json(self, json_obj):
        logger.log(LOGLEVEL_TRACE, json.dumps(json_obj, sort_keys=True, separators=(',', ':'), indent=4))

    def _get_sistr_predictions(self, sistr_analysis_href):
        sistr_pred_json = None

        analysis = self.irida_connector.get(sistr_analysis_href)

        sistr_href = self._get_rel_from_links('outputFile/sistr-predictions', analysis['links'])
        sistr_pred = self.irida_connector.get_file(sistr_href)
        try:
            sistr_pred_json = sistr_pred.json()
        except json.decoder.JSONDecodeError as e:
            raise SistrResultsException("Error parsing JSON") from e

        if (sistr_pred_json is None):
            raise SistrResultsException("Could not get SISTR predictions for sistr " + sistr_analysis_href)

        self._log_json(sistr_pred_json)

        return sistr_pred_json

    def _has_sample_in_paired(self, paired_json):
        if (len(paired_json) > 0 and paired_json[0]['links'] is not None):
            return self._has_rel_in_links('sample', paired_json[0]['links'])
        else:
            return False

    def _get_sample_from_paired(self, paired_json):
        sample = None

        links = paired_json[0]['links']
        sample_href = self._get_rel_from_links('sample', links)
        sample = self.irida_connector.get(sample_href)

        if sample is None:
            raise Exception("Could not get sample corresponding to " + paired_path)
        else:
            return sample

    def get_user_project(self, project_id):
        """
        Gets information on a particular IRIDA project
        :param project_id: The id of the project to search.
        :return:  The JSON results for the IRIDA project
        """
        return self.irida_connector.get('/api/projects/' + str(project_id))

    def get_user_projects(self):
        """
        Gets information on all projects in IRIDA the current user has access to.

        :return:  The JSON results for all the user-accessible projects in IRIDA.
        """
        return self.irida_connector.get_resources('/api/projects')

    def get_sistr_results_for_project(self, project, sistr_workflow_ids, sample_created_date_min):
        """
        Gets information on all SISTR results in a project (that is, automated SISTR results).
        This is structured as a list of objects (on per sample) containing a SampleSistrInfo object if available and whether
        or not these results are available for a sample (from keys 'has_results').

        :param project:  The project to search through.
        :param sistr_workflow_ids: A list of SISTR workflow ids, None for all workflow results.
        :param sample_created_date_min: Only return results for samples created after this date.
        :return: A list of dictionaries containing SampleSistrInfo objects and/or indicators for missing results.
        """
        sistr_results = []

        sistr_results_for_project = self.irida_connector.get_resources('/api/projects/' + str(project) + '/samples')

        for sample in sistr_results_for_project:
            if self._created_since(sample, sample_created_date_min):
                logger.debug("sample [id=%s, name=%s, createdDate=%s] created before %s, skipping.",
                             sample['identifier'], sample['sampleName'], sample['createdDate'],
                             sample_created_date_min)
                continue

            sample_pairs = self.irida_connector.get_resources('/api/samples/' + sample['identifier'] + '/pairs')

            sistr_info = SampleSistrInfo.create_empty_info(sample)

            if len(sample_pairs) > 0:
                for sequencing_object in sample_pairs:
                    if (self._has_rel_in_links('analysis/sistr', sequencing_object['links'])):
                        sistr_rel = self._get_rel_from_links('analysis/sistr', sequencing_object['links'])
                        sistr = self.irida_connector.get(sistr_rel)

                        if (sistr['analysisState'] != 'COMPLETED'):
                            logger.debug(
                                "Skipping automated SISTR results associated with sample=%s as state is not 'COMPLETED'.",
                                sample['identifier'])
                        else:
                            new_sistr_info = self.get_sistr_info_from_submission(sistr)
                            sistr_info = self._update_sample_sistr_info(sistr_info, new_sistr_info,
                                                                        sistr_workflow_ids)

                sistr_results.append(sistr_info)
        return sistr_results

    def _created_since(self, sample, sample_created_date_min):
        return sample_created_date_min and datetime.fromtimestamp(
            sample['createdDate'] / 1000) < sample_created_date_min

    def _update_sample_sistr_info(self, curr_sistr_info, new_sistr_info, sistr_workflow_ids):
        if sistr_workflow_ids is None or new_sistr_info.get_submission_workflow_id() in sistr_workflow_ids:
            if not curr_sistr_info.has_sistr_results():
                return new_sistr_info
            elif new_sistr_info.get_qc_status() == 'PASS' and curr_sistr_info.get_qc_status() != 'PASS':
                return new_sistr_info
            elif new_sistr_info.get_qc_status() == 'PASS' and (
                    new_sistr_info.get_submission_created_date() > curr_sistr_info.get_submission_created_date()):
                return new_sistr_info
        else:
            logger.debug(
                "Skipping sistr submission [id=%s, workflowId=%s]. workflowId not in %s".format(
                    new_sistr_info.get_sample_id(), new_sistr_info.get_submission_workflow_id(), sistr_workflow_ids))

        return curr_sistr_info

    def get_sistr_info_from_submission(self, submission):
        """
        Gets the relevent SISTR information from an IRIDA SISTR AnalysisSubmission.

        :param submission: The IRIDA SISTR AnalysisSubmission data structure.
        :return: The SISTR information object (SampleSistrInfo) fro this submission.
        """
        sistr_info = {}

        links = submission['links']
        paired_path = self._get_rel_from_links('input/paired', links)
        sistr_analysis_href = self._get_rel_from_links('analysis', links)

        unpaired_path = self._get_rel_from_links('input/unpaired', links)
        unpaired_files = self.irida_connector.get_resources(unpaired_path)

        if len(unpaired_files) > 0:
            self_href = self._get_rel_from_links('self', submission['links'])
            raise Exception(
                'Error: unpaired files were found for analysis submission ' + self_href + '. SISTR results from unpaired files not currently supported')

        paired = self.irida_connector.get_resources(paired_path)

        sistr_info['paired_files'] = paired
        sistr_info['sistr_predictions'] = self._get_sistr_predictions(sistr_analysis_href)
        sistr_info['has_results'] = True
        if (self._has_sample_in_paired(paired)):
            sistr_info['sample'] = self._get_sample_from_paired(paired)
        else:
            sistr_info['sample'] = None
        sistr_info['submission'] = submission

        return SampleSistrInfo(sistr_info)

    def get_sistr_submissions_for_user(self, sistr_workflow_ids=None):
        """
        Gets all SISTR results accessible by a user (includes those submitted by a user or shared with a user via a project).
        :param sistr_workflow_ids: A list of SISTR workflow ids, None for getting all workflow results.
        :return:  A list of SampleSistrInfo objects that the current user has access to.
        """
        sistr_submissions_for_user = self.irida_connector.get_resources('/api/analysisSubmissions/analysisType/sistr')

        sistr_analysis_list = []
        for sistr in sistr_submissions_for_user:
            try:
                if (sistr['analysisState'] == 'COMPLETED'):
                    if sistr_workflow_ids is None or sistr['workflowId'] in sistr_workflow_ids:
                        sistr_analysis_list.append(self.get_sistr_info_from_submission(sistr))
                    else:
                        logger.debug("Skipping sistr submission [id=%s, workflowId=%s]. workflowId not in %s".format(
                            sistr['identifier'], sistr['workflowId'], sistr_workflow_ids))
                else:
                    logger.debug('Skipping incompleted sistr submission [id=%s]', sistr['identifier'])
            except (HTTPError, SistrResultsException) as e:
                logger.warning('Could not read information for SISTR analysis submission id=' + str(
                    sistr['identifier']) + ', name=' + str(sistr['name']) + ', ignoring these results. ' + str(e))

        return sistr_analysis_list

    def get_sistr_submissions_shared_to_project(self, project_id, sistr_workflow_ids=None):
        """
        Gets all SISTR results that were shared to a project.
        :param project_id: The project identifier.
        :param sistr_workflow_ids: A list of SISTR workflow ids of results to include.
        :return: A list of all SISTR results shared to the project.
        """
        sistr_submissions_for_project = self.irida_connector.get_resources(
            '/api/projects/' + str(project_id) + '/analyses/sistr')

        sistr_analysis_list = []
        for sistr in sistr_submissions_for_project:
            try:
                if (sistr['analysisState'] == 'COMPLETED'):
                    if sistr_workflow_ids is None or sistr['workflowId'] in sistr_workflow_ids:
                        sistr_analysis = self._get_sistr_submission(sistr['identifier'])
                        sistr_analysis_list.append(self.get_sistr_info_from_submission(sistr_analysis))
                    else:
                        logger.debug("Skipping sistr submission [id=%s, workflowId=%s]. workflowId not in %s".format(
                            sistr['identifier'], sistr['workflowId'], sistr_workflow_ids))
                else:
                    logger.debug('Skipping incompleted sistr submission [id=%s]', sistr['identifier'])
            except (HTTPError, SistrResultsException) as e:
                logger.warning('Could not read information for SISTR analysis submission id=' + str(
                    sistr['identifier']) + ', name=' + str(sistr['name']) + ', ignoring these results. ' + str(e))

        return sistr_analysis_list

    def _get_sistr_submission(self, id):
        """
        Gets the particular AnalysisSubmission for the given SISTR analysis id.
        :param id: The SISTR analysis id.
        :return: The AnalysisSubmission data for the SISTR id.
        """
        return self.irida_connector.get('/api/analysisSubmissions/' + str(id))



import json
import logging
from urllib.parse import urlsplit

from rauth import OAuth2Service

LOGLEVEL_TRACE = 5

logger = logging.getLogger("irida_connector")

def decode_access_token(s):
    """
    Used to decode the access token.
    :param s: The input bytes.
    :return: A dictionary.
    """
    return json.loads(s.decode('utf-8'))


class IridaConnector(object):
    """Low-level connections to the IRIDA REST API"""

    def __init__(self, client_id, client_secret, username, password, base_url, timeout):
        """
        Creates a new object for connecting to the IRIDA REST API

        :param client_id:  The client_id for IRIDA.
        :param client_secret:  The client secret for IRIDA.
        :param username:  The username of the user for IRIDA.
        :param password:  The password for the user.
        :param base_url:  The base URL for IRIDA (minis the '/api' part)
        :param timeout:  The maximum timeout for any connection to IRIDA.

        :return: An object which can be used to connect to IRIDA.
        """
        base_url = base_url.rstrip('/')
        self._base_path = urlsplit(base_url).path
        self._timeout = timeout

        access_token_url = base_url + '/api/oauth/token'

        params = {
            "data": {
                "grant_type": "password",
                "client_id": client_id,
                "client_secret": client_secret,
                "username": username,
                "password": password
            }
        }

        oauth_service = OAuth2Service(
            client_id=client_id,
            client_secret=client_secret,
            name="irida",
            access_token_url=access_token_url,
            base_url=base_url
        )

        token = oauth_service.get_access_token(decoder=decode_access_token, **params)
        self.session = oauth_service.get_session(token)

    def get(self, path):
        """
        A GET request to a particular path in IRIDA.

        :param path: The path to GET, minus the IRIDA url (e.g., '/projects').

        :return:  The result of rauth.OAuth2Service.get()
        """
        path = self._join_path(path)
        logger.debug("Getting path=" + path)
        response = self.session.get(path, timeout=self._timeout)

        if (response.ok):
            self._log_json(response.json())
            return response.json()['resource']
        else:
            response.raise_for_status()

    def _join_path(self, path):
        if (self._base_path is None or self._base_path == ''):
            return path
        # If passed full URL like http://localhost/path, don't add on base_path
        elif (urlsplit(path).scheme != ''):
            return path
        else:
            if (path[0] == '/'):
                return self._base_path + path
            else:
                return self._base_path + '/' + path

    def get_resources(self, path):
        """
        GETs the resources from an IRIDA REST API endpoint (e.g., (get '/projects')['resources']

        :param path: The path to GET the resources.

        :return:  The ['resources'] part of the GET JSON response.
        """
        return self.get(path)['resources']

    def get_file(self, path):
        """
        GETs the file contents from an IRIDA REST API endpoint.

        :param path: The path to GET the file.

        :return:  The file contents.
        """
        return self.session.get(path, headers={'Accept': 'text/plain'}, timeout=self._timeout)

    def _log_json(self, json_obj):
        logger.log(LOGLEVEL_TRACE, json.dumps(json_obj, sort_keys=True, separators=(',', ':'), indent=4))



import uuid


class IridaSistrWorkflow:
    """Functionality for conversion between workflow ids and versions."""

    WORKFLOW_IDS = {
        '0.1': 'e559af58-a560-4bbd-997e-808bfbe026e2',
        '0.1.0': 'e559af58-a560-4bbd-997e-808bfbe026e2',
        '0.2': 'e8f9cc61-3264-48c6-81d9-02d9e84bccc7',
        '0.2.0': 'e8f9cc61-3264-48c6-81d9-02d9e84bccc7',
        '0.3': '92ecf046-ee09-4271-b849-7a82625d6b60',
        '0.3.0': '92ecf046-ee09-4271-b849-7a82625d6b60',
        None: None
    }
    WORKFLOW_VERSIONS = {
        'e559af58-a560-4bbd-997e-808bfbe026e2': '0.1',
        'e8f9cc61-3264-48c6-81d9-02d9e84bccc7': '0.2',
        '92ecf046-ee09-4271-b849-7a82625d6b60': '0.3',
        None: None
    }

    @classmethod
    def workflow_id_to_version(cls, workflow_id):
        """
        Converts an IRIDA workflow id to a version number.
        :param workflow_id: The workflow id.
        :return: The version number for the workflow.
        """
        return cls.WORKFLOW_VERSIONS[workflow_id]

    @classmethod
    def workflow_version_to_id(cls, workflow_version):
        """
        Converts an IRIDA workflow version to an id.
        :param workflow_version: The workflow version.
        :return: The id of the workflow.
        """
        return cls.WORKFLOW_IDS[workflow_version]

    @classmethod
    def workflow_versions_to_ids(cls, workflow_versions):
        """
        Converts a list of IRIDA workflow versions to a list of ids.
        :param workflow_versions: The list of workflow versions.
        :return: The list of ids.
        """
        return None if workflow_versions is None else [cls.workflow_version_to_id(x) for x in workflow_versions]

    @classmethod
    def workflow_ids_or_versions_to_ids(cls, workflow_versions_or_ids):
        """
        Converts a list of IRIDA workflow versions or ids to a list of ids.
        :param workflow_versions: The list of workflow versions or ids.
        :return: The list of workflow ids.
        """
        if workflow_versions_or_ids is None:
            return None
        else:
            ids_list = []
            for value in workflow_versions_or_ids:
                try:
                    id = uuid.UUID('{' + value + '}')
                    ids_list.append(str(id))
                except ValueError:
                    id = cls.workflow_version_to_id(value)
                    ids_list.append(id)

            return ids_list

    @classmethod
    def all_workflow_versions(cls):
        """
        Gets a list of all workflow versions.
        :return: A list of all workflow versions.
        """
        versions = list(cls.WORKFLOW_VERSIONS.values())
        versions.remove(None)
        return versions



from datetime import datetime

from irida_sistr_results.irida_sistr_workflow import IridaSistrWorkflow


class SampleSistrInfo(object):
    """Stores and provides access to SISTR/IRIDA information for a particular sample."""

    def __init__(self, sistr_info):
        self.sistr_info = sistr_info

    @classmethod
    def create_empty_info(cls, sample, sequencing_object=None):
        """
        Creates an empty SampleSistrInfo object (no SISTR results).
        :param sample: The sample this object refers to.
        :param sequencing_object: The sequencing object (default None).
        :return: An empty SampleSistrInfo.
        """
        data = {'sample': sample,
                'has_results': False
                }

        if sequencing_object is not None:
            data['paired_files'] = sequencing_object

        return cls(data)

    def has_sistr_results(self):
        return self.sistr_info['has_results']

    def get_sample_name(self):
        if (self.sistr_info['sample'] is not None):
            return self.sistr_info['sample']['sampleName']
        else:
            return "N/A"

    def _get_sistr(self):
        return self.sistr_info['sistr_predictions'][0]

    def get_serovar(self):
        return self._get_sistr()['serovar']

    def get_serovar_antigen(self):
        return self._get_sistr()['serovar_antigen']

    def get_serovar_cgmlst(self):
        serovar_cgmlst = self._get_sistr()['serovar_cgmlst']
        if (serovar_cgmlst is None):
            return ''
        else:
            return serovar_cgmlst

    def get_serogroup(self):
        return self._get_sistr()['serogroup']

    def get_h1(self):
        return self._get_sistr()['h1']

    def get_h2(self):
        return self._get_sistr()['h2']

    def get_o_antigen(self):
        return self._get_sistr()['o_antigen']

    def get_qc_status(self):
        return 'MISSING' if (not self.has_sistr_results()) else self._get_sistr()['qc_status']

    def get_qc_status_numerical(self):
        """Gets numerical value for QC status, used for sorting"""
        return ['MISSING', 'FAIL', 'WARNING', 'PASS'].index(self.get_qc_status())

    def get_qc_messages(self):
        return self._get_sistr()['qc_messages']

    def get_cgmlst_subspecies(self):
        return self._get_sistr()['cgmlst_subspecies']

    def get_cgmlst_genome(self):
        return self._get_sistr()['cgmlst_genome_match']

    def get_cgmlst_matching_alleles(self):
        return self._get_sistr()['cgmlst_matching_alleles']

    def get_cgmlst_matching_total_alleles(self):
        return str(self.get_cgmlst_matching_alleles()) + '/330'

    def get_cgmlst_matching_proportion(self):
        return float(self.get_cgmlst_matching_alleles()) / 330

    def get_cgmlst_sequence_type(self):
        return self._get_sistr()['cgmlst_ST']

    def get_mash_subspecies(self):
        return self._get_sistr()['mash_subspecies']

    def get_mash_serovar(self):
        return self._get_sistr()['mash_serovar']

    def get_mash_genome(self):
        return self._get_sistr()['mash_genome']

    def get_mash_distance(self):
        return self._get_sistr()['mash_distance']

    def get_submission_url(self, irida_base_url):
        submission_url = irida_base_url
        if submission_url.endswith('/'):
            submission_url += 'analysis/'
        else:
            submission_url += '/analysis/'
        submission_url += self.get_submission_identifier()

        return submission_url

    def get_submission_identifier(self):
        return self.sistr_info['submission']['identifier']

    def get_submission_workflow_id(self):
        return self.sistr_info['submission']['workflowId']

    def get_submission_workflow_version(self):
        return IridaSistrWorkflow.workflow_id_to_version(self.get_submission_workflow_id())

    def get_submission_created_date(self):
        return datetime.fromtimestamp(self.sistr_info['submission']['createdDate'] / 1000)

    def get_sample_created_date(self):
        if (self.sistr_info['sample'] is not None):
            return datetime.fromtimestamp(self.sistr_info['sample']['createdDate'] / 1000)
        else:
            return "N/A"

    def get_sample_id(self):
        if (self.sistr_info['sample'] is not None):
            return self.sistr_info['sample']['identifier']
        else:
            return "N/A"

    def get_paired_id(self):
        return self.sistr_info['paired_files'][0]['identifier']



import abc
import csv
from collections import OrderedDict
from datetime import datetime
from operator import methodcaller

import xlsxwriter

from irida_sistr_results.version import __version__


class SistrResultsWriter(object):
    """Abstract class resonsible for writing SISTR results to a table format"""

    def __init__(self, irida_url, appname, username, sample_created_min_date=None):
        """
        Construct a new SistrResultsWriter object corresponding to the passed irida_url

        :param irida_url: The URL to the IRIDA instance, used to insert URLs into the table
        :param appname: The application name.
        :param username: The name of the user generating these results.
        :param sample_created_min_date: The minimum date for including samples.
        """
        __metaclass__ = abc.ABCMeta
        self.irida_url = irida_url
        self.appname = appname
        self.username = username
        self.row = 0
        self.end_of_project = False
        self.sample_created_min_date = sample_created_min_date

    @abc.abstractmethod
    def _write_row(self, row):
        """Abstract method for writing a particular row for the table"""
        return

    @abc.abstractmethod
    def _write_header(self, header):
        """Abstract method for writing the table header"""
        return

    def _formatting(self):
        """Override to implement additional formatting to the table after all rows have finished writing"""
        return

    def _set_end_of_project(self, end_of_project):
        """Sets whether or not we are at the end row of a project group"""
        self.end_of_project = end_of_project

    def _is_end_of_project(self):
        return self.end_of_project

    def _finish(self):
        return

    def close(self):
        """Closes writing to a file"""
        return

    def get_row(self):
        """Gets the current row number"""
        return self.row

    def set_row(self, row):
        """
        Sets the current row number

        :param row: The new row number
        """
        self.row = row

    def _get_header_list(self):
        """
        Get a list of header titles for the table

        :return: A list of header titles.
        """
        return [
            'Project ID',
            'Sample Name',
            'QC Status',
            'Serovar (overall)',
            'Serovar (antigen)',
            'Serovar (cgMLST)',
            'Serogroup',
            'H1',
            'H2',
            'O-antigen',
            'cgMLST Subspecies',
            'cgMLST Matching Genome',
            'Alleles Matching Genome',
            'cgMLST Percent Matching',
            'cgMLST Sequence Type',
            'Mash Subspecies',
            'Mash Serovar',
            'Mash Matching Genome Name',
            'Mash Distance',
            'QC Messages',
            'IRIDA URL',
            'Sample Created Date',
            'IRIDA Sample Identifier',
            'IRIDA File Pair Identifier',
            'IRIDA Submission Identifier',
            'IRIDA Analysis Date',
            'IRIDA Workflow Version',
            'IRIDA Workflow ID',
        ]

    def _format_timestamp(self, timestamp):
        return timestamp.isoformat(sep=' ')

    def _get_header_index(self, title):
        """
        Gets the particular index from the headers given the title.

        :param title: The title of the header column.

        :return: The index into the header list.
        """
        return self._get_header_list().index(title)

    def _get_row_list(self, project, result):
        """
        Given the project number and result object, creates a list of relavent information to print per row.

        :param project: The current project identifier.
        :param result:  The current SistrInfo result object.

        :return: A list of relevant information for the row
        """
        return [
            project,
            result.get_sample_name(),
            result.get_qc_status(),
            result.get_serovar(),
            result.get_serovar_antigen(),
            result.get_serovar_cgmlst(),
            result.get_serogroup(),
            result.get_h1(),
            result.get_h2(),
            result.get_o_antigen(),
            result.get_cgmlst_subspecies(),
            result.get_cgmlst_genome(),
            result.get_cgmlst_matching_alleles(),
            result.get_cgmlst_matching_proportion(),
            result.get_cgmlst_sequence_type(),
            result.get_mash_subspecies(),
            result.get_mash_serovar(),
            result.get_mash_genome(),
            result.get_mash_distance(),
            result.get_qc_messages(),
            result.get_submission_url(self.irida_url),
            result.get_sample_created_date(),
            result.get_sample_id(),
            result.get_paired_id(),
            result.get_submission_identifier(),
            result.get_submission_created_date(),
            result.get_submission_workflow_version(),
            result.get_submission_workflow_id(),
        ]

    def _get_no_results_row_list(self, project, result):
        """
        Gets a list respresenting no/missing results for a sample.

        :param project: The current project identifier.
        :param result:  The current SistrInfo result object.

        :return: A list of relevant information in the case of a no/missing result row.
        """
        return [
            project,
            result.get_sample_name(),
            result.get_qc_status(),
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            result.get_sample_created_date(),
            result.get_sample_id(),
            None,
            None,
            None,
            None,
            None,
        ]

    def _get_irida_sistr_run_info(self):
        """
        Gets information about the run of this irida-sistr-results script.
        :return: Information in the form of a Dict of key/value pairs.
        """
        info = OrderedDict()
        info['appname'] = self.appname
        info['version'] = __version__
        info['irida_url'] = self.irida_url
        info['username'] = self.username
        info['app_run_date'] = datetime.now()

        if self.sample_created_min_date:
            info['sample_created_min_date'] = self.sample_created_min_date

        return info

    def write(self, sistr_results):
        """
        Writes out the results to an appropriate file with the appropriate format

        :param sistr_results:  The SISTR results to write to a table.
        :return: None
        """
        self.set_row(0)
        self._write_header(self._get_header_list())
        self.set_row(1)

        for project in sorted(sistr_results.keys(), key=int):
            row_start_project = self.get_row()

            sistr_results_project = sistr_results[project]

            sistr_results_sorted = sorted(sistr_results_project.values(), key=methodcaller('get_sample_name'))
            sistr_results_sorted = sorted(sistr_results_sorted, key=methodcaller('get_qc_status_numerical'),
                                          reverse=True)
            for index, result in enumerate(sistr_results_sorted):
                # last element in this list
                if (index == len(sistr_results_sorted) - 1):
                    self._set_end_of_project(True)

                if (not result.has_sistr_results()):
                    self._write_row(self._get_no_results_row_list(project, result))
                else:
                    self._write_row(self._get_row_list(project, result))
                self.set_row(self.get_row() + 1)

            self._set_end_of_project(False)

        self._formatting()
        self._finish()


class SistrCsvWriter(SistrResultsWriter):
    """An abstact writer used to create CSV/tab-delimited files"""

    def __init__(self, irida_url, appname, username, out_file, sample_created_min_date=None):
        super(SistrCsvWriter, self).__init__(irida_url, appname, username, sample_created_min_date)
        out_file_h = open(out_file, 'w')
        self.writer = csv.writer(out_file_h, delimiter="\t", quotechar='"', quoting=csv.QUOTE_MINIMAL)

    def _write_header(self, header):
        run_info = self._get_irida_sistr_run_info()
        sample_create_msg = ' using only samples created after ' + self.sample_created_min_date.isoformat(
            sep=' ') if 'sample_created_min_date' in run_info else ''
        self._write_row([
            "# Results generated from {} version={} connecting to IRIDA={} as user={} on date={} {}".format(
                run_info['appname'],
                run_info['version'],
                run_info['irida_url'],
                run_info['username'],
                run_info['app_run_date'],
                sample_create_msg)])
        self.writer.writerow(header)

    def _write_row(self, row):
        self.writer.writerow(row)

    def _get_row_list(self, project, result):
        return [
            project,
            result.get_sample_name(),
            result.get_qc_status(),
            result.get_serovar(),
            result.get_serovar_antigen(),
            result.get_serovar_cgmlst(),
            result.get_serogroup(),
            result.get_h1(),
            result.get_h2(),
            result.get_o_antigen(),
            result.get_cgmlst_subspecies(),
            result.get_cgmlst_genome(),
            result.get_cgmlst_matching_alleles(),
            "{0:.1f}".format(result.get_cgmlst_matching_proportion() * 100) + '%',
            result.get_cgmlst_sequence_type(),
            result.get_mash_subspecies(),
            result.get_mash_serovar(),
            result.get_mash_genome(),
            result.get_mash_distance(),
            result.get_qc_messages(),
            result.get_submission_url(self.irida_url),
            result.get_sample_created_date(),
            result.get_sample_id(),
            result.get_paired_id(),
            result.get_submission_identifier(),
            result.get_submission_created_date(),
            result.get_submission_workflow_version(),
            result.get_submission_workflow_id(),
        ]


class SistrExcelWriter(SistrResultsWriter):
    """A writer object for writing SISTR results to an excel spreadsheet"""

    def __init__(self, irida_url, appname, username, out_file, sample_created_min_date=None):
        super(SistrExcelWriter, self).__init__(irida_url, appname, username, sample_created_min_date)
        self.workbook = xlsxwriter.Workbook(out_file, {'default_date_format': 'yyyy/mm/dd'})
        self.worksheet = self.workbook.add_worksheet('Data')
        self.index_of_cgmlst_percent = self._get_header_list().index('cgMLST Percent Matching')
        self.index_of_date_formats = [self._get_header_list().index('Sample Created Date'),
                                      self._get_header_list().index('IRIDA Analysis Date')
                                      ]
        self.percent_format = self.workbook.add_format({'num_format': '0.0%'})

    def _get_header_column_number(self, title):
        """
        Gets the particular column number from the headers given the title.

        :param title: The title of the header column.

        :return: The column number (starting with 1) from the header list.
        """
        return self._get_header_index(title) + 1

    def _get_header_column_letter(self, title):
        """
        Gets the particular column letter from the headers given the title.

        :param title: The title of the header column.

        :return: The column letter (starting with A) from the header list.
        """
        return self._to_letter(self._get_header_index(title))

    def _range_stitle(self, title):
        """
        Gets the particular column letter range from the headers given a single title.

        :param title: The title of the header column.

        :return: The column range (e.g., A:A) from the header list.
        """
        return self._range_title(title, title)

    def _range_title(self, start_title, end_title):
        scol = self._get_header_index(start_title)
        ecol = self._get_header_index(end_title)
        return self._to_range_col_1(scol, ecol)

    def _to_letter(self, col):
        return chr(ord('A') + col)

    def _to_range_col(self, start_col, end_col):
        return self._to_letter(start_col) + ':' + self._to_letter(end_col)

    def _to_range_col_1(self, start_col, end_col):
        return self._to_letter(start_col) + '1:' + self._to_letter(end_col) + '1'

    def _to_range_row(self, start_title, start_row, end_row):
        start_col = self._get_header_column_letter(start_title)
        return start_col + str(start_row) + ':' + start_col + str(end_row)

    def _to_range(self, start_row, end_row, start_col, end_col):
        return self._to_letter(start_col) + str(start_row) + ':' + self._to_letter(end_col) + str(end_row)

    def _write_header(self, header):
        merged_header_format = self.workbook.add_format()
        merged_header_format.set_bold()
        merged_header_format.set_align('center')

        header_format = self.workbook.add_format()
        header_format.set_bold()

        col = 0
        for item in header:
            self.worksheet.write(self.get_row(), col, item, header_format)
            col += 1

        self.worksheet.set_column(self._range_stitle('Project ID'), 15)
        self.worksheet.set_column(self._range_title('Sample Name', 'Serogroup'), 20)
        self.worksheet.set_column(self._range_stitle('H1'), 10)
        self.worksheet.set_column(self._range_stitle('H2'), 10)
        self.worksheet.set_column(self._range_title('O-antigen', 'cgMLST Subspecies'), 20)
        self.worksheet.set_column(self._range_title('cgMLST Matching Genome', 'cgMLST Sequence Type'), 25)
        self.worksheet.set_column(self._range_title('Mash Subspecies', 'Mash Serovar'), 20)
        self.worksheet.set_column(self._range_title('Mash Matching Genome Name', 'IRIDA Analysis Date'), 30)

    def _write_row(self, row):
        col = 0
        for item in row:
            if col == self.index_of_cgmlst_percent:
                self.worksheet.write(self.get_row(), col, item, self._get_percent_format())
            elif col in self.index_of_date_formats:
                self.worksheet.write(self.get_row(), col, item, self._get_date_format())
            else:
                self.worksheet.write(self.get_row(), col, item, self._get_regular_format())
            col += 1

        if self._is_end_of_project():
            bottom_format = self.workbook.add_format({'bottom': 5})
            self.worksheet.set_row(self.get_row(), None, bottom_format)

    def _get_percent_format(self):
        if (self._is_end_of_project()):
            return self.workbook.add_format({'num_format': '0.0%', 'bottom': 5})
        else:
            return self.workbook.add_format({'num_format': '0.0%'})

    def _get_date_format(self):
        if (self._is_end_of_project()):
            return self.workbook.add_format({'num_format': 'yyyy/mm/dd', 'bottom': 5})
        else:
            return self._get_default_date_format()

    def _get_default_date_format(self):
        return self.workbook.add_format({'num_format': 'yyyy/mm/dd'})

    def _get_regular_format(self):
        if (self._is_end_of_project()):
            return self.workbook.add_format({'bottom': 5})
        else:
            return self.workbook.add_format()

    def _formatting(self):
        format_pass = self.workbook.add_format({'bg_color': '#DFF0D8'})
        format_warning = self.workbook.add_format({'bg_color': '#FCF8E3'})
        format_fail = self.workbook.add_format({'bg_color': '#F2DEDE'})
        format_missing = self.workbook.add_format({'bg_color': '#BBBBBB'})
        form_range = self._to_range_row('QC Status', 1, self.get_row())
        self.worksheet.conditional_format(form_range, {'type': 'cell',
                                                       'criteria': '==',
                                                       'value': '"PASS"',
                                                       'format': format_pass})
        self.worksheet.conditional_format(form_range, {'type': 'cell',
                                                       'criteria': '==',
                                                       'value': '"WARNING"',
                                                       'format': format_warning})
        self.worksheet.conditional_format(form_range, {'type': 'cell',
                                                       'criteria': '==',
                                                       'value': '"FAIL"',
                                                       'format': format_fail})
        self.worksheet.conditional_format(form_range, {'type': 'cell',
                                                       'criteria': '==',
                                                       'value': '"MISSING"',
                                                       'format': format_missing})
        self.worksheet.freeze_panes(1, 3)

    def _finish(self):
        info_worksheet = self.workbook.add_worksheet('Settings')
        info = self._get_irida_sistr_run_info()

        date_format = self._get_default_date_format()

        row = 0
        for k, v in info.items():
            info_worksheet.write(row, 0, k)

            if k == 'app_run_date' or k == 'sample_created_min_date':
                info_worksheet.write(row, 1, v, date_format)
            else:
                info_worksheet.write(row, 1, v)
            info_worksheet.write(row, 1, v)
            row += 1

        bold_format = self.workbook.add_format()
        bold_format.set_bold()

        info_worksheet.set_column('A:A', 25, bold_format)
        info_worksheet.set_column('B:B', 20)

    def close(self):
        """Closes the workbook"""
        self.workbook.close()



import logging

from irida_sistr_results.irida_sistr_workflow import IridaSistrWorkflow

logger = logging.getLogger("irida-sistr-results")


class IridaSistrResults(object):
    """Class for constructing the top-level data structures mapping projects to lists of SISTR results."""

    def __init__(self, irida_api, include_user_results, update_existing_with_user_results,
                 sistr_workflow_versions_or_ids=None, sample_created_min_date=None):
        """
        Creates a new IridaSistrResults object.

        :param irida_api:  The IridaAPI object for connecting to IRIDA.
        :param include_user_results:  Whether or not to include all user-accessible results (or just automated SISTR results).
        :param update_existing_with_user_results:  Whether or not to update existing results with newer results run by a user.
        :param sistr_workflow_versions_or_ids: A list of SISTR workflow versions (or UUIDs) of results to include.
        :param sample_created_min_date: The minimum (oldest) sample created date for samples to include results for.

        :return:  A new IridaSistrResults object.
        """
        self.irida_api = irida_api
        self.include_user_results = include_user_results
        self.update_existing_with_user_results = update_existing_with_user_results
        self.sistr_workflow_ids = IridaSistrWorkflow.workflow_ids_or_versions_to_ids(sistr_workflow_versions_or_ids)
        self.sistr_results = {}
        self.sample_project = {}
        self.sample_created_min_date = sample_created_min_date

    def get_sistr_results_all_projects(self):
        """
        Gets SISTR results from all projects accessible by the current user.

        :return:  All SISTR results from all projects.
        """
        projects = self.irida_api.get_user_projects()
        return self._get_sistr_results(projects)

    def get_sistr_results_from_projects(self, project_ids):
        """
        Gets SISTR results from the list of IRIDA project identifiers.

        :param project_ids:  The list of project ids.

        :return:  All SISTR results from the passed projects.
        """
        projects = []
        for p in project_ids:
            project = self.irida_api.get_user_project(p)
            projects.append(project)

        return self._get_sistr_results(projects)

    def _get_sistr_results(self, projects):
        for p in projects:
            logger.debug("Working on project [" + p['identifier'] + ', ' + p['name'] + ']')
            self._load_sistr_results_for_project(p)
            self._load_sistr_results_shared_to_project(p['identifier'])

        if (self.include_user_results):
            self._load_sistr_results_from_user()

        return self.sistr_results

    def _load_sistr_results_from_user(self):
        user_results = self.irida_api.get_sistr_submissions_for_user(self.sistr_workflow_ids)
        self._load_additional_sistr_results(user_results)

    def _load_sistr_results_shared_to_project(self, project_id):
        project_results = self.irida_api.get_sistr_submissions_shared_to_project(project_id, self.sistr_workflow_ids)
        self._load_additional_sistr_results(project_results)

    def _load_additional_sistr_results(self, additional_results):
        for result in additional_results:
            sample_id = result.get_sample_id()

            if (sample_id in self.sample_project and result.has_sistr_results()):
                for project in self.sample_project[sample_id]:
                    sistr_results_project = self.sistr_results[project]
                    if sample_id in sistr_results_project:
                        if (not sistr_results_project[sample_id].has_sistr_results()):
                            sistr_results_project[sample_id] = result
                        elif (sistr_results_project[
                                  sample_id].get_submission_created_date() < result.get_submission_created_date()):
                            if (self.update_existing_with_user_results):
                                logger.debug(self._result_to_sample_log_string(sistr_results_project[sample_id], result,
                                                                               "older") + " Updating.")
                                sistr_results_project[sample_id] = result
                            else:
                                logger.debug(
                                    "Found result for sample=" + result.get_sample_name() + ". Will not replace with exisiting result.")
                        else:
                            logger.debug(self._result_to_sample_log_string(sistr_results_project[sample_id], result,
                                                                           "newer") + " Not updating.")

    def _load_sistr_results_for_project(self, project):
        project_id = project['identifier']

        if (project_id in self.sistr_results):
            raise Exception("Error: project " + str(project_id) + " already examined")

        self.sistr_results[project_id] = {}
        project_results = self.irida_api.get_sistr_results_for_project(project_id, self.sistr_workflow_ids,
                                                                       self.sample_created_min_date)

        for result in project_results:
            if result is None:
                logger.warning("None result found in project " + str(project_id) + ", will skip")
            else:
                sample_id = result.get_sample_id()
                self.sistr_results[project_id][sample_id] = result
                if (sample_id in self.sample_project):
                    self.sample_project[sample_id].append(project_id)
                else:
                    self.sample_project[sample_id] = [project_id]

    def _timef(self, timestamp):
        return timestamp.isoformat(sep=' ')

    def _result_to_sample_log_string(self, r1, r2, word):
        return "Sample [name=" + r1.get_sample_name() + ", id=" + str(
            r1.get_sample_id()) + "] has exisiting analysis [id=" + r1.get_submission_identifier() + ", created_date=" + self._timef(
            r1.get_submission_created_date()) + "] " + word + " than analysis [id=" + r2.get_submission_identifier() + ", created_date=" + self._timef(
            r2.get_submission_created_date()) + "]."
